__all__ = [
    "apigateway",
    "certificatemanager",
    "cloudformation",
    "cloudwatch",
    "dynamodb",
    "ec2",
    "elasticsearch",
    "events",
    "iam",
    "kinesis",
    "kinesisfirehose",
    "kms",
    "awslambda",
    "logs",
    "redshift",
    "resourcegroups",
    "route53",
    "s3",
    "secretsmanager",
    "sns",
    "sqs",
    "ssm",
    "stepfunctions",
]
